"""This module contains classes designed to read information from an OPeNDAP
DMR or NetCDF-4 file. These classes will group the input into science
variables, metadata, coordinates, dimensions and ancillary data sets.

"""

from __future__ import annotations

from abc import abstractmethod
from typing import Union
import re
import xml.etree.ElementTree as ET

from netCDF4 import Variable as NetCDF4Variable

from varinfo.attribute_container import (
    AttributeContainerBase,
    AttributeContainerFromDmr,
    AttributeContainerFromNetCDF4,
)
from varinfo.cf_config import CFConfig
from varinfo.utilities import CF_REFERENCE_ATTRIBUTES


InputVariableType = Union[ET.Element, NetCDF4Variable]


class VariableBase(AttributeContainerBase):
    """A class to represent a single variable contained within a granule
    representation. It will produce an object in which references are
    fully qualified, and also augmented by any overrides from the supplied
    configuration file.

    """

    def __init__(
        self,
        variable: InputVariableType,
        cf_config: CFConfig,
        namespace: str,
        full_name_path: str,
    ):
        """Extract the references contained within the appropriate
        CF-Convention attributes of the variable. These should be augmented
        by information from the `CFConfig` instance passed to the class.

        Additionally, store all metadata attributes in a dictionary.

        """
        super().__init__(variable, cf_config, namespace, full_name_path)
        self.group_path, self.name = self._extract_group_and_name()
        self.data_type = self._get_data_type(variable)
        self.references = self._get_all_cf_references()
        self.dimensions = self._extract_dimensions(variable)
        self.shape = self._get_shape(variable)

    @abstractmethod
    def _get_data_type(self, variable: InputVariableType):
        """Extract a string representation of the variable data type."""

    @abstractmethod
    def _get_shape(self, variable: InputVariableType):
        """Extract the shape of the variable array."""

    @abstractmethod
    def _get_raw_dimensions(self, variable: InputVariableType):
        """Retrieve the dimension names as they are stored within the
        variable.

        """

    def get_range(self) -> list[float] | None:
        """Retrieve the range of valid data from the variable metadata. First,
        try to parse the `valid_range` metadata attribute. If this is
        absent, check for a combination of `valid_min` and `valid_max`.

        If insufficient range information is present in the metadata, this
        method will return `None`.

        """
        valid_range = self.attributes.get('valid_range')

        if valid_range is None:
            valid_min = self.attributes.get('valid_min')
            valid_max = self.attributes.get('valid_max')

            if valid_min is not None and valid_max is not None:
                valid_range = [valid_min, valid_max]

        return valid_range

    def get_valid_min(self) -> float | None:
        """Retrieve the minimum valid value for variable data from the
        associated metadata. First try to retrieve data from the
        `valid_min` metadata attribute. If this is absent, then try to
        retrieve the same information from the `valid_range` metadata.

        If insufficient range information is present in the metadata, this
        method will return `None`.

        """
        valid_min = self.attributes.get('valid_min')

        if valid_min is None:
            valid_range = self.attributes.get('valid_range')
            if isinstance(valid_range, list) and len(valid_range) == 2:
                valid_min = valid_range[0]

        return valid_min

    def get_valid_max(self) -> float | None:
        """Retrieve the maximum valid value for variable data from the
        associated metadata. First try to retrieve data from the
        `valid_max` metadata attribute. If this is absent, then try to
        retrieve the same information from the `valid_range` metadata.

        If insufficient range information is present in the metadata, this
        method will return `None`.

        """
        valid_max = self.attributes.get('valid_max')

        if valid_max is None:
            valid_range = self.attributes.get('valid_range')
            if isinstance(valid_range, list) and len(valid_range) == 2:
                valid_max = valid_range[1]

        return valid_max

    def get_references(self) -> set[str]:
        """Combine the references extracted from the ancillary_variables,
        coordinates and dimensions data into a single set for VarInfo to
        use directly.

        The variable dimensions are cast as a set to allow combination with
        the other set attributes of the `VariableBase` class. The
        dimensions attribute is kept as a list prior to combination in the
        full set of variable references to ensure that the ordering of the
        dimensions is preserved.

        """
        return set(self.dimensions).union(*self.references.values())

    def is_geographic(self) -> bool:
        """Use heuristics to determine if the variable is a geographic
        coordinate based on its units. A latitude variable will have units
        'degrees_north' and a longitude variable with have units
        'degrees_east'.

        """
        return self.is_longitude() or self.is_latitude()

    def is_latitude(self) -> bool:
        """Determine if the variable is a latitude based on the `units`
        metadata attribute being 'degrees_north' or other similar options
        as defined in section 4.1 of the CF Conventions (v1.8).

        """
        return self.attributes.get('units') in [
            'degrees_north',
            'degree_north',
            'degrees_N',
            'degree_N',
            'degreesN',
            'degreeN',
        ]

    def is_longitude(self) -> bool:
        """Determine if the variable is a longitude based on the `units`
        metadata attribute being 'degrees_east' or other similar options
        as defined in section 4.2 of the CF Conventions (v1.8).

        """
        return self.attributes.get('units') in [
            'degrees_east',
            'degree_east',
            'degrees_E',
            'degree_E',
            'degreesE',
            'degreeE',
        ]

    def is_projection_x_or_y(self) -> bool:
        """Determine if the variable is a projected x or y horizontal spatial
        coordinate based on the `standard_name` metadata attribute being
        `projection_x_coordinate` or `projection_y_coordinate`, as defined
        by the CF Conventions (v1.9). Note, geostationary projections have
        coordinates with `standard_name` metadata attribute values of
        `projection_x_angular_coordinate` and
        `projection_y_angular_coordinate`.

        """
        return self.is_projection_x() or self.is_projection_y()

    def is_projection_x(self) -> bool:
        """Determine if the variable is a projected x spatial coordinate based
        on the `standard_name` metadata attribute being
        `projection_x_coordinate`, as defined by the CF Conventions
        (v1.9). Note, geostationary projections have coordinate with
        `standard_name` metadata attribute value of
        `projection_x_angular_coordinate`.

        """
        return self.attributes.get('standard_name') in [
            'projection_x_coordinate',
            'projection_x_angular_coordinate',
        ]

    def is_projection_y(self) -> bool:
        """Determine if the variable is a projected y spatial coordinate based
        on the `standard_name` metadata attribute being
        `projection_y_coordinate`, as defined by the CF Conventions
        (v1.9). Note, geostationary projections have coordinate with
        `standard_name` metadata attribute value of
        `projection_y_angular_coordinate`.

        """
        return self.attributes.get('standard_name') in [
            'projection_y_coordinate',
            'projection_y_angular_coordinate',
        ]

    def is_temporal(self) -> bool:
        """Determine if the variable is a time based on the `units`
        metadata attribute being 'since' or other similar options
        as defined in section 4.4 of the CF Conventions (v1.8).

        """
        return ' since ' in self.attributes.get('units', '')

    def _get_all_cf_references(self) -> dict[str, set[str]]:
        """Retrieve a dictionary containing all CF-Convention attributes
        within the variable that have references to other variables in the
        granule. These variable references will be fully qualified paths.

        """
        return {
            attribute_name: self._get_cf_references(attribute_name)
            for attribute_name in CF_REFERENCE_ATTRIBUTES
            if attribute_name in self.attributes
        }

    def _get_cf_references(self, attribute_name: str) -> set[str]:
        """Retrieve an attribute from the parsed variable metadata, correcting
        for any known artefacts (missing or incorrect references). Then
        split this string and qualify the individual references.

        """
        return self._extract_references(self.attributes.get(attribute_name))

    def _extract_references(self, attribute_string: str) -> set[str]:
        """Given a string value of an attribute, which may contain multiple
        references to dataset, split that string based on either commas,
        or spaces (or both together). Then if any reference is a relative
        path, make it absolute.

        """
        if attribute_string is not None:
            raw_references = re.split(r'\s+|,\s*', attribute_string)
            references = set(self._qualify_references(raw_references))
        else:
            references = set()

        return references

    def _extract_dimensions(self, variable: ET.Element) -> list[str]:
        """Find the dimensions for the variable in question. If there are
        overriding dimensions from the `earthdata-varinfo` configuration
        file, these are used instead of the raw dimensions from the `.dmr`. All
        references are converted to absolute paths in the granule. A set of all
        fully qualified references is returned.

        """
        dimensions_override = self.metadata_overrides.get('dimensions')

        if dimensions_override is not None:
            dimensions = re.split(r'\s+|,\s*', dimensions_override)
        else:
            dimensions = [
                dimension
                for dimension in self._get_raw_dimensions(variable)
                if dimension is not None
            ]

        return self._qualify_references(dimensions)

    def _qualify_references(self, raw_references: list[str]) -> list[str]:
        """Take a list of local references to other variables, and produce a
        list of absolute references.

        Trailing colons are removed from variable references. These might
        occur in some CF-Convention defined formats of the grid_mapping
        metadata attribute. E.g. "crs: grid_y crs: grid_x" (See section 5.6
        of CF-Conventions).

        """
        references = []

        if self.group_path is not None:
            for reference in raw_references:
                reference = reference.rstrip(':')

                if reference.startswith('../'):
                    # Reference is relative, and requires manipulation
                    absolute_path = self._construct_absolute_path(reference)
                elif reference.startswith('/'):
                    # Reference is already absolute
                    absolute_path = reference
                elif reference.startswith('./'):
                    # Reference is in the same group as this variable
                    absolute_path = self.group_path + reference[1:]
                else:
                    # Reference is in the same group as this variable
                    absolute_path = '/'.join([self.group_path, reference])

                references.append(absolute_path)

        else:
            for reference in raw_references:
                reference = reference.rstrip(':')

                if reference.startswith('/'):
                    absolute_path = reference
                else:
                    absolute_path = f'/{reference}'

                references.append(absolute_path)

        return references

    def _construct_absolute_path(self, reference: str) -> str:
        """For a relative reference to another variable (e.g. '../latitude'),
        construct an absolute path by combining the reference with the
        group path of the variable.

        """
        relative_prefix = '../'
        group_path_pieces = self.group_path.split('/')

        while reference.startswith(relative_prefix):
            reference = reference[len(relative_prefix) :]
            group_path_pieces.pop()

        absolute_path = group_path_pieces + [reference]
        return '/'.join(absolute_path)

    def _extract_group_and_name(self) -> tuple[str, str]:
        """Extract the group and base name of a variable from the full path,
        e.g. '/this/is/my/variable' should return a two-element tuple:
        ('/this/is/my', 'variable').

        """
        split_full_path = self.full_name_path.split('/')
        name = split_full_path.pop(-1)
        group_path = '/'.join(split_full_path) or None

        return group_path, name


class VariableFromDmr(VariableBase, AttributeContainerFromDmr):
    """This child class inherits from the `VariableBase` class, and implements
    the abstract methods assuming the variable source is part of an XML
    element tree.

    """

    def __init__(
        self,
        element: ET.Element,
        cf_config: CFConfig,
        namespace: str,
        full_name_path: str,
        all_dimensions_sizes: dict[str, int],
    ):
        self.all_dimensions_sizes = all_dimensions_sizes
        super().__init__(element, cf_config, namespace, full_name_path)

    def _get_data_type(self, variable: ET.Element) -> str:
        """Extract a string representation of the variable data type."""
        return variable.tag.lstrip(self.namespace).lower()

    def _get_shape(self, variable: ET.Element) -> tuple[int]:
        """Extract the shape of the variable data array. First explore
        XML for - DIM child elements with a size
        attribute - true for HDF5 files and anonymous dimensions
        (note no name). If found (with or without names), use them. If
        not found, use all_dimensions_sizes dictionary with matching dimension
        names to define the variable shape. Note the all_dimensions_sizes
        dictionary is filled when parsing xml for variables, capturing
        dimensions names and sizes for dimension entries - as found in
        NetCDF files with named dimensions.
        """

        # Retrieve the shape from the dimension size in the
        # <Dim size=xx/> element.
        shape = [
            int(dim_size.attrib['size'])
            for dim_size in variable.findall(f'.//{self.namespace}Dim[@size]')
        ]

        # Retrieve the shape from dimension size in the
        # <Dimension size=xx/> element
        if len(shape) != len(self.dimensions):
            for dim_name in self.dimensions:
                if dim_name in self.all_dimensions_sizes:
                    index = self.dimensions.index(dim_name)
                    shape.insert(index, self.all_dimensions_sizes[dim_name])

        return shape

    def _get_raw_dimensions(self, variable: ET.Element) -> list[str]:
        """Extract the raw dimension names from a <Dim /> XML element."""
        return [
            dimension.get('name')
            for dimension in variable.findall(f'{self.namespace}Dim')
        ]


class VariableFromNetCDF4(VariableBase, AttributeContainerFromNetCDF4):
    """This child class inherits from the `VariableBase` class, and implements
    the abstract methods assuming the variable source is part of a NetCDF-4
    file.

    """

    def _get_data_type(self, variable: NetCDF4Variable) -> str:
        """Extract a string representation of the variable data type.

        * First try `variable.datatype.name`. This can be `None` for some
          string type variables.
        * If `variable.datatype.name` is `None`, try `variable.datatype.dtype`.

        """
        return variable.datatype.name or variable.datatype.dtype.__name__

    def _get_shape(self, variable: NetCDF4Variable) -> tuple[int]:
        """Extract the shape of the variable data array."""
        return variable.shape

    def _get_raw_dimensions(self, variable: NetCDF4Variable) -> list[str]:
        """Retrieve the dimension names as they are stored within the
        variable.

        """
        return list(variable.dimensions)



"""This script leverages the `python-cmr` library to conduct a CMR search
and get a granule download URL. With the granule download URL and the
`requests` library a granule is downloaded via https and saved locally.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal
import os.path

from cmr import GranuleQuery, CMR_OPS, CMR_SIT, CMR_UAT
import requests

from varinfo.exceptions import (
    CMRQueryException,
    MissingGranuleDownloadLinks,
    MissingPositionalArguments,
    GranuleDownloadException,
    DirectoryCreationException,
    GetEdlTokenException,
)


CmrEnvType = Literal[CMR_OPS, CMR_UAT, CMR_SIT]
urs_token_endpoints = {
    CMR_OPS: 'https://urs.earthdata.nasa.gov/api/nams/edl_user_token',
    CMR_UAT: 'https://uat.urs.earthdata.nasa.gov/api/nams/edl_user_token',
    CMR_SIT: 'https://sit.urs.earthdata.nasa.gov/api/nams/edl_user_token',
}


def get_granules(
    concept_id: str | None = None,
    collection_shortname: str | None = None,
    collection_version: str | None = None,
    provider: str | None = None,
    cmr_env: CmrEnvType = CMR_OPS,
    auth_header: str | None = None,
) -> Sequence:
    """Search CMR to retrieve granules for a specific collection given:

    * concept_id: a CMR collection or granule concept ID. For most
      workflows, including the generateVariableDrafts within the CMR
      GraphQL interface, this is anticipated to be a collection concept ID,
      but the `GranuleQuery` object can also make queries with a granule
      concept IDs.
    * shortname/short_name: e.g., 'SNDRSNIML2CCPRET' or 'M2T1NXSLV'
    * collection_version/version: a collection version: '2' or 5.12.4'
    * provider: data center ID (e.g. GES_DISC, PODAAC, POCLOUD, EEDTEST)
    * cmr_env/mode: CMR environments (OPS, UAT, and SIT)
    * auth_header: Authorization HTTP header, either:
      - A header with a LaunchPad token: 'Authorization: <token>'
      - A header with an EDL bearer token: 'Authorization: Bearer <token>'

    For a successful search response concept_id or short_name, version and
    provider must be entered along with a bearer_token.

    """
    granule_query = GranuleQuery(mode=cmr_env)

    # Ensure the Authorization header was provided for CMR authentication
    if auth_header is not None:
        granule_query.headers = {'Authorization': auth_header}
    else:
        raise MissingPositionalArguments('auth_header')

    if concept_id is not None:
        # If a collection or granule concept ID is specified, can query CMR
        # just using that parameter.
        query_parameters = {'concept_id': concept_id}
    elif collection_shortname and collection_version and provider is not None:
        # Otherwise use the combination of short_name, version, and provider
        query_parameters = {
            'short_name': collection_shortname,
            'version': collection_version,
            'provider': provider,
        }
    else:
        # If neither condition is met, there aren't enough CMR query parameters
        # to identify the collection.
        raise MissingPositionalArguments(
            'concept_id or collection_shortname, collection_version, and provider'
        )

    # Assign parameters to GranuleQuery
    granule_query.parameters(
        downloadable=True, sort_key='-start_date', **query_parameters
    )
    try:
        # .get() is the number of links to return (e.g. page_size)
        granule_response = granule_query.get(10)

    except Exception as cmr_exception:
        # Capture exception raised due to CMR failure and return an exception
        # with a more user-friendly message.
        raise CMRQueryException(str(cmr_exception)) from cmr_exception

    if len(granule_response) == 0:
        # No granules were identified, so no files can be downloaded and parsed
        raise IndexError(
            'No granules were found with selected parameters and user permissions'
        )

    return granule_response


def get_granule_link(granule_response: Sequence) -> str:
    """Get the granule download link from CMR."""
    granule_link = next(
        (
            link['href']
            for link in granule_response[0].get('links', [])
            if link['rel'].endswith('/data#') and 'inherited' not in link
        ),
        None,
    )

    if granule_link is None:
        raise MissingGranuleDownloadLinks(granule_response)

    return granule_link


def download_granule(
    granule_link: str, auth_header: str, out_directory: str = os.getcwd()
) -> str:
    """Use the requests module to download data via https.
    * granule_link: granule download URL.
    * auth_header: Authorization HTTP header, either:
      - A header with a LaunchPad token: 'Authorization: <token>'
      - A header with an EDL bearer token: 'Authorization: Bearer <token>'
    * out_directory: path to save downloaded granule
        (the default is the current directory).
    """
    # Create `out_directory` if it does not exist and create out_filename
    if not os.path.isdir(out_directory):
        try:
            os.mkdir(out_directory)
        except Exception as os_exception:
            raise DirectoryCreationException(str(os_exception)) from os_exception

    out_filename = os.path.join(out_directory, os.path.basename(granule_link))

    try:
        # Write content of data to out_filename and return response
        response = requests.get(
            granule_link, headers={'Authorization': auth_header}, timeout=10
        )

        with open(out_filename, 'wb') as file_download:
            file_download.write(response.content)

        return out_filename
    except Exception as requests_exception:
        # Custom exception for error from `requests.get`
        raise GranuleDownloadException(str(requests_exception)) from requests_exception


def get_edl_token_from_launchpad(
    launchpad_token: str, cmr_env: CmrEnvType
) -> str | None:
    """Retrieve an EDL token given a LaunchPad token.
    * launchpad_token: A LaunchPad token with no header prefixes:
      <Launchpad token>
    * cmr_env/mode: CMR environments (OPS, UAT, and SIT)
    """
    url_urs_endpoint = urs_token_endpoints[cmr_env]
    try:
        response = requests.post(
            url=url_urs_endpoint, data=f'token={launchpad_token}', timeout=10
        )
        response.raise_for_status()

    except Exception as requests_exception:
        raise GetEdlTokenException(str(requests_exception)) from requests_exception

    return response.json().get('access_token')


def get_edl_token_header(auth_header: str, cmr_env: CmrEnvType) -> str:
    """Helper function for getting the header for an EDL token.
    * auth_header: Authorization HTTP header, either:
      - A header with a LaunchPad token: 'Authorization: <token>'
      - A header with an EDL bearer token: 'Authorization: Bearer <token>'
    * cmr_env/mode: CMR environments (OPS, UAT, and SIT)
    """
    if 'Bearer ' not in auth_header:
        edl_auth_header = f'Bearer {get_edl_token_from_launchpad(auth_header, cmr_env)}'
    else:
        edl_auth_header = auth_header
    return edl_auth_header


def get_dmr_xml_url(granule_response: Sequence) -> str:
    """Gets the OPeNDAP url from a CMR granule response and appends `.dmr.xml`"""
    granule_link = next(
        (
            link['href']
            for link in granule_response[0].get('links', [])
            if link['rel'].endswith('/service#')
            and 'inherited' not in link
            and 'opendap' in link['href']
        ),
        None,
    )

    if granule_link is None:
        raise MissingGranuleDownloadLinks(granule_response)

    return granule_link + '.dmr.xml'



"""This module contains lower-level functionality that can be abstracted into
functions. Primarily this improves readability of the source code, and
allows finer-grained unit testing of each smaller part of functionality.

For CF-Convention reference attributes see the following sections of the
Conventions:

- Section 3.4: ancillary_variables
- Sections 4 and 5: coordinates
- Section 5.6: grid_mapping
- Section 7.1: bounds
- Section 7.2: cell_measure
- Section 7.5: geometry, interior_ring, node_coordinates, node_count,
               nodes, part_node_count

susbset_control_variables is a new attribute introduced by the Data
Services team, primarily for use with references for segmented data.

"""

from __future__ import annotations

from typing import Any
from xml.etree.ElementTree import Element
import functools
import re

from netCDF4 import Dataset as NetCDF4Dataset
import numpy as np

from varinfo.exceptions import DmrNamespaceError


CF_REFERENCE_ATTRIBUTES = [
    'ancillary_variables',
    'bounds',
    'cell_measures',
    'coordinates',
    'geometry',
    'grid_mapping',
    'interior_ring',
    'node_coordinates',
    'node_count',
    'nodes',
    'part_node_count',
    'subset_control_variables',
]

DAP4_TO_NUMPY_MAP = {
    'Char': np.uint8,
    'Byte': np.uint8,
    'Int8': np.int8,
    'UInt8': np.uint8,
    'Int16': np.int16,
    'UInt16': np.uint16,
    'Int32': np.int32,
    'UInt32': np.uint32,
    'Int64': np.int64,
    'UInt64': np.uint64,
    'Float32': np.float32,
    'Float64': np.float64,
    'String': str,
    'URL': str,
}


def recursive_get(input_dictionary: dict, keys: list[str]):
    """Extract a value from an arbitrarily nested dictionary."""
    try:
        nested_value = functools.reduce(dict.get, keys, input_dictionary)
    except TypeError:
        # This catches when there is a missing intermediate key
        nested_value = None

    return nested_value


def split_attribute_path(full_path: str) -> list[str]:
    """Take the full path to a metadata attribute and return the list of
    keys that locate that attribute within the global attributes.
    This function can account for the input path to having, or omitting, a
    leading '/' character.

    """
    return full_path.lstrip('/').split('/')


def get_xml_namespace(root_element: Element) -> str:
    """Given the root element of an XML document, extract the associated
    namespace. This allows for the full qualification of child elements.
    The root element of a dmr file is expected to be a Dataset tag.

    """
    match = re.match('(.+)Dataset', root_element.tag)

    if match:
        xml_namespace = match.groups()[0]
    else:
        raise DmrNamespaceError(root_element.tag)

    return xml_namespace


def get_xml_attribute(
    variable: Element,
    attribute_name: str,
    namespace: str,
    default_value: Any | None = None,
) -> Any | None:
    """Extract the value of an XML Attribute tag from a `.dmr`. First search
    the supplied variable element for a fully qualified Attribute child
    element, with a name property matching the requested attribute name. If
    there is no matching tag, return the `default_value`, which can be
    user-defined, or default to `None`. If present, the returned value is
    cast as the type indicated by the Attribute tag's `type` property.

    Attributes with multiple Value children will return a list of all those
    children, cast as the indicated type. Attributes that are containers of
    nested attributes will return a dictionary structure.

    """
    attribute_element = variable.find(
        f'{namespace}Attribute' f'[@name="{attribute_name}"]'
    )

    if attribute_element is not None:
        value_type = attribute_element.get('type', 'String')

        if value_type != 'Container':
            attribute_value = get_xml_attribute_value(
                attribute_element,
                namespace,
                value_type,
                default_value,
            )
        else:
            attribute_value = get_xml_container_attribute(attribute_element, namespace)

    else:
        attribute_value = default_value

    return attribute_value


def get_xml_attribute_value(
    attribute_element: Element,
    namespace: str,
    value_type: str,
    default_value: Any | None = None,
) -> Any | None:
    """Extract the value (single or list) for an XML attribute. If there are
    no attributes matching the required name, then return the supplied default
    value. If no default value is supplied, the default used is `None`.

    """
    numpy_type = DAP4_TO_NUMPY_MAP.get(value_type, str)

    value_elements = attribute_element.findall(f'{namespace}Value')

    if len(value_elements) > 1:
        attribute_value = [
            numpy_type(value_element.text) for value_element in value_elements
        ]
    elif len(value_elements) == 1:
        attribute_value = numpy_type(value_elements[0].text)
    else:
        attribute_value = default_value

    return attribute_value


def get_xml_container_attribute(
    container_element: Element, namespace: str
) -> dict[str, Any | None]:
    """Extract a dictionary of attribute values when an attribute is a container
    for further attributes. This function is recursive, and so nested containers
    will be treated in the same way.

    """
    attribute_dictionary = {}

    for child in list(container_element):
        child_name = child.get('name')
        child_type = child.get('type', 'String')

        if child_type != 'Container':
            attribute_dictionary[child_name] = get_xml_attribute_value(
                child,
                namespace,
                child_type,
            )
        else:
            attribute_dictionary[child_name] = get_xml_container_attribute(
                child,
                namespace,
            )

    return attribute_dictionary


def get_full_path_xml_attribute(
    dmr_document: Element,
    attribute_path: str,
    namespace: str,
) -> Any | None:
    """Helper function that retrieves the value of an XML attribute, given the
    full path to that attribute. If the XML attribute is not present, then
    `None` is returned.

    """
    attribute_element = dmr_document

    try:
        for path_part in attribute_path.lstrip('/').split('/')[:-1]:
            attribute_element = attribute_element.find(f'.//*[@name="{path_part}"]')

        attribute_value = get_xml_attribute(
            attribute_element,
            attribute_path.split('/')[-1],
            namespace,
        )
    except AttributeError:
        attribute_value = None

    return attribute_value


def get_full_path_netcdf4_attribute(
    netcdf_dataset: NetCDF4Dataset,
    attribute_path: str,
) -> Any | None:
    """Helper function that retrieves the value of a metadata attribute from a
    NetCDF-4 file, given the full path to that attribute. If the metadata
    attribute is not present, then `None` is returned.

    """
    container_path = '/'.join(attribute_path.split('/')[:-1])
    attribute_name = attribute_path.split('/')[-1]

    try:
        if container_path != '':
            attribute_container = netcdf_dataset[container_path]
        else:
            attribute_container = netcdf_dataset

        attribute_value = attribute_container.getncattr(attribute_name)
    except (KeyError, IndexError, AttributeError):
        attribute_value = None

    return attribute_value



"""This module contains classes that represent groups (e.g., containers of
variables within a NetCDF-4 file or OPeNDAP DMR). A group has metadata
attributes and child variables.

"""

from __future__ import annotations

from abc import abstractmethod
from typing import Union
import xml.etree.ElementTree as ET

from netCDF4 import Group as NetCDF4Group

from varinfo.attribute_container import (
    AttributeContainerBase,
    AttributeContainerFromDmr,
    AttributeContainerFromNetCDF4,
)
from varinfo.cf_config import CFConfig
from varinfo.utilities import DAP4_TO_NUMPY_MAP


InputGroupType = Union[ET.Element, NetCDF4Group]


class GroupBase(AttributeContainerBase):
    """A class to represent a single group contained within a granule
    representation. It will produce an object with attributes and a set of
    fully qualified variables within the group.

    """

    def __init__(
        self,
        group: InputGroupType,
        cf_config: CFConfig,
        namespace: str,
        full_name_path: str,
    ):
        """First extract all metadata attributes on the group, accounting for
        overrides defined in the CFConfig file. Then parse the paths of all
        child variables in the group.

        """
        super().__init__(group, cf_config, namespace, full_name_path)
        self.variables = self._parse_variables(group)

    @abstractmethod
    def _parse_variables(self, group: InputGroupType) -> set[str]:
        """An abstract method to retrieve a set of all variables that are
        direct children of the group.

        """


class GroupFromDmr(GroupBase, AttributeContainerFromDmr):
    """This child class inherits from the `GroupBase` class and implements the
    abstract methods assuming the group source is a Dataset Metadata Response
    (DMR) XML document retrieved from OPeNDAP.

    """

    def _parse_variables(self, group: ET.Element) -> set[str]:
        """Returns full paths of all child variables in the group."""
        return {
            '/'.join([self.full_name_path.rstrip('/'), child.get('name', '')])
            for child in group
            if child.tag.replace(self.namespace, '') in DAP4_TO_NUMPY_MAP
        }


class GroupFromNetCDF4(GroupBase, AttributeContainerFromNetCDF4):
    """This child class inherits from the `GroupBase` class and implements the
    abstract methods assuming the group source is a NetCDF-4 file.

    """

    def _parse_variables(self, group: NetCDF4Group) -> set[str]:
        """Returns full paths of all child variables in the group."""
        return {
            '/'.join([self.full_name_path.rstrip('/'), variable])
            for variable in group.variables
        }



"""This module contains a class designed to read information from a `.dmr`
file. This should group the input into science variables, metadata,
coordinates, dimensions and ancillary data sets.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from os.path import exists
from typing import Any, Union
import json
import re
import xml.etree.ElementTree as ET

from netCDF4 import Dataset, Group

from varinfo.cf_config import CFConfig
from varinfo.exceptions import (
    InvalidConfigFileFormatError,
    MissingConfigurationFileError,
)
from varinfo.group import GroupFromDmr, GroupFromNetCDF4
from varinfo.utilities import (
    DAP4_TO_NUMPY_MAP,
    get_full_path_netcdf4_attribute,
    get_full_path_xml_attribute,
    get_xml_namespace,
)
from varinfo.variable import VariableFromDmr, VariableFromNetCDF4


DimensionsGroupType = dict[tuple[str], set[str]]
OutputGroupType = Union[GroupFromDmr, GroupFromNetCDF4]
OutputVariableType = Union[VariableFromDmr, VariableFromNetCDF4]


class VarInfoBase(ABC):
    """An abstract base class to represent the full dataset of a granule,
    having reading information from a representation of that granule. Currently
    supported granule representations: OPeNDAP Dataset Metadata Response (DMR),
    NetCDF-4 file.

    """

    def __init__(
        self,
        file_path: str,
        short_name: str | None = None,
        config_file: str | None = None,
    ):
        """Distinguish between variables containing references to other
        datasets, and those that do not. The former are considered science
        variables, providing they are not considered coordinates or
        dimensions for another variable.

        Each variable contains references to their specific coordinates and
        dimensions, allowing the retrieval of all required variables for a
        specified list of science variables.

        """
        self.config_file = config_file
        self.short_name = short_name
        self.mission = None
        self.namespace = None
        self.groups: dict[str, OutputGroupType] = {}
        self.variables: dict[str, OutputVariableType] = {}
        self.references: set[str] = set()
        self.metadata: dict[str, OutputVariableType] = {}
        self.all_dimensions_sizes: dict[str, int] = {}

        self._set_var_info_config()
        self._read_dataset(file_path)
        self._set_mission_and_short_name()
        self.cf_config = self._set_cf_config()
        self._extract_variables()

    @abstractmethod
    def _read_dataset(self, file_path: str):
        """This method parses a file at the specified location using
        functionality specific to the type of input (e.g. a `.dmr` file).

        """

    @abstractmethod
    def _extract_variables(self):
        """Iterate through all variables in the retrieved dataset. For each
        variable create an instance of a `Variable` (using the relevant
        child class), and assign it to either the `metadata_variables`
        or the `variable_with_coordinates` dictionary accordingly.

        """

    def _assign_variable(self, variable_object):
        """Save the `Variable` instance in the dictionary containing all
        variables. Additionally, the set of references for all variables is
        updated.

        """
        self.references.update(variable_object.get_references())
        self.variables[variable_object.full_name_path] = variable_object

    def _set_var_info_config(self):
        """Read the VarInfo configuration JSON file, containing locations to
        search for the collection short_name attribute, and the mapping
        from short_name to satellite mission.

        """
        if self.config_file is not None and not exists(self.config_file):
            raise MissingConfigurationFileError(self.config_file)
        elif self.config_file is not None and self.config_file.endswith('.json'):
            with open(self.config_file, 'r', encoding='utf-8') as file_handler:
                self.var_info_config = json.load(file_handler)
        elif self.config_file is not None:
            raise InvalidConfigFileFormatError(self.config_file)
        else:
            self.var_info_config = {}

    def _set_cf_config(self) -> CFConfig:
        """Instantiate a CFConfig object, to contain any rules for exclusions,
        required fields and augmentations to CF attributes that are not
        contained within a granule from the specified collection.

        """
        return CFConfig(self.mission, self.short_name, self.config_file)

    def _set_mission_and_short_name(self):
        """Check a series of potential locations for the collection short name
        of the granule. Once that is determined, match that short name to its
        associated mission.

        """
        if self.short_name is None:
            self._set_short_name()

        if self.short_name is not None:
            self.mission = next(
                (
                    name
                    for pattern, name in self.var_info_config.get('Mission', {}).items()
                    if re.match(pattern, self.short_name) is not None
                ),
                None,
            )

    @abstractmethod
    def _set_short_name(self):
        """Iterate through the locations in the data granule representation to
        find the first value entry for the collections short name in the granule
        metadata.

        """

    def get_variable(self, variable_path: str) -> OutputVariableType | None:
        """Retrieve a variable specified by an absolute path. First check the
        variables with coordinates, before checking those without. If there
        are no matching variables, a value of `None` is returned.

        """
        return self.variables.get(variable_path)

    def get_all_variables(self) -> set[str]:
        """Retrieve a set of names for all variables in the granule."""
        return set(self.variables.keys())

    def get_variables_with_coordinates(self) -> dict[str, OutputVariableType]:
        """Return only variables with a `coordinates` metadata attribute.
        This list excludes any variables listed as an excluded science
        variable in the configuration file supplied to the object.

        """
        exclusion_pattern = re.compile(
            '|'.join(self.cf_config.excluded_science_variables)
        )

        return {
            variable_path: variable
            for variable_path, variable in self.variables.items()
            if variable.references.get('coordinates') is not None
            and not self.variable_is_excluded(variable, exclusion_pattern)
        }

    def _is_spatial_temporal_dimension(self, dimension_path: str) -> bool:
        """A helper method simplifying the list comprehension in
        `is_science_variable`.
        """
        dimension = self.get_variable(dimension_path)
        return dimension is not None and (
            dimension.is_geographic()
            or dimension.is_temporal()
            or dimension.is_projection_x_or_y()
        )

    def is_science_variable(self, variable: OutputVariableType) -> bool:
        """Determine if a variable is a science variable.
        A science variable is classified as a variable that is NOT
        a spatial temporal dimension variable. But contains
        geographic, temporal, and or projected spatial dimensions, or
        a coordinate or grid mapping reference attribute.

        """
        if any(
            self._is_spatial_temporal_dimension(dimension)
            for dimension in variable.dimensions
            if dimension != variable.full_name_path
            and not variable.full_name_path.endswith('_bnds')
        ):
            return True

        if (
            variable.references.get('coordinates') is not None
            or variable.references.get('grid_mapping') is not None
        ):
            return True

        return False

    def get_science_variables(self) -> set[str]:
        """Retrieve a set of names for all variables that have coordinate
        references, that are not themselves used as dimensions, coordinates
        or ancillary date for another variable.

        """
        exclusions_pattern = re.compile(
            '|'.join(self.cf_config.excluded_science_variables)
        )

        filtered_with_coordinates = {
            variable_path
            for variable_path, variable in self.variables.items()
            if variable_path is not None
            and self.is_science_variable(variable)
            and not self.variable_is_excluded(variable_path, exclusions_pattern)
        }

        return filtered_with_coordinates - self.references

    def get_metadata_variables(self) -> set[str]:
        """Retrieve set of names for all variables that do no have
        coordinates references, that are not themselves used as dimensions,
        coordinates, ancillary data for another variable, or are
        science variables.

        Additionally, any excluded science variables, that are contained
        in the variables class attribute should be considered a metadata
        variable.

        """
        exclusions_pattern = re.compile(
            '|'.join(self.cf_config.excluded_science_variables)
        )

        non_coordinate_variables = {
            variable_path
            for variable_path, variable in self.variables.items()
            if variable_path is not None
            and (
                not self.variable_is_excluded(variable_path, exclusions_pattern)
                and not self.is_science_variable(variable)
                and not variable.full_name_path.endswith('_bnds')
            )
        }

        return non_coordinate_variables - self.references

    @staticmethod
    def variable_is_excluded(
        variable_name: str, exclusions_pattern: re.Pattern
    ) -> bool:
        """Ensure the variable name does not match any collection specific
        exclusion rules.

        """
        if exclusions_pattern.pattern != '':
            exclude_variable = exclusions_pattern.match(variable_name) is not None
        else:
            exclude_variable = False

        return exclude_variable

    def get_required_variables(self, requested_variables: set[str]) -> set[str]:
        """Retrieve requested variables and recursively search for all
        associated dimension and coordinate variables. The returned set
        should be the union of the science variables, coordinates and
        dimensions.

        The requested variables are also augmented to include required
        variables for the collection, as indicated by the CFConfig class
        instance, and any references within those variables.

        All variables included within the `CFConfig.required_variables` instance
        attribute will be included in any request for required variables. There
        is no pattern matching applied from the `ApplicabilityType` for the
        items in the `RequiredVariables` section of the configuration file.

        """
        if self.cf_config.required_variables:
            cf_required_pattern = re.compile(
                '|'.join(self.cf_config.required_variables)
            )

            cf_required_variables = {
                variable
                for variable in self.get_all_variables()
                if variable is not None and re.match(cf_required_pattern, variable)
            }
        else:
            cf_required_variables = set()

        requested_variables.update(cf_required_variables)
        required_variables: set[str] = set()

        while len(requested_variables) > 0:
            variable_name = requested_variables.pop()
            required_variables.add(variable_name)

            variable = self.get_variable(variable_name)

            if variable is not None:
                # Add variable. Enqueue references not already present in
                # required set. (Also checking that they are real variables,
                # and not non-variable dimensions)
                variable_references = {
                    reference_variable
                    for reference_variable in variable.get_references()
                    if self.get_variable(reference_variable) is not None
                }
                requested_variables.update(
                    variable_references.difference(required_variables)
                )

        return self.exclude_fake_dimensions(required_variables)

    def get_required_dimensions(self, variables: set[str]) -> set[str]:
        """Return a single set of all variables that are used as dimensions
        for any of the listed variables.

        """
        return set(
            dimension
            for variable in variables
            for dimension in getattr(self.get_variable(variable), 'dimensions', [])
            if self.get_variable(dimension) is not None
        )

    def get_missing_variable_attributes(self, variable_name: str) -> dict[str, Any]:
        """Return a dictionary of all attributes for a variable that is not present
        in the granule file (e.g. grid_mapping variable in collections that are not
        fully CF compliant). The metadata attributes and the overrides for the
        variables would need to be in the configuration file.

        """
        return self.cf_config.get_metadata_overrides(variable_name)

    def get_references_for_attribute(
        self, list_of_variables: list[str], reference_attribute_name: str
    ) -> set[str]:
        """Return a single set of all references in a specific metadata attribute
        for a list of variables (e.g. bounds, coordinates, cf attributes). The full
        list of supported metadata attributes can be found in
        varinfo.utilities::CF_REFERENCE_ATTRIBUTES

        Iterate through all requested variables and extract a list of
        references for the metadata attribute. This will produce a list of lists,
        which should be flattened into a single list and then combined into a set
        to remove duplicates.

        """
        reference_set = [
            self.get_variable(variable).references.get(reference_attribute_name)
            for variable in list_of_variables
            if self.get_variable(variable).references.get(reference_attribute_name)
            is not None
        ]
        return set(
            variable_reference
            for variable_references in reference_set
            for variable_reference in variable_references
        )

    def get_spatial_dimensions(self, variables: set[str]) -> set[str]:
        """Return a single set of all variables that are both used as
        dimensions for any of the input variables, and that are horizontal
        spatial dimensions (either geographic or projected).

        """
        return set().union(
            self.get_geographic_spatial_dimensions(variables),
            self.get_projected_spatial_dimensions(variables),
        )

    def get_geographic_spatial_dimensions(self, variables: set[str]) -> set[str]:
        """Return a single set of all the variables that are both used as
        dimensions for any of the input variables, and that are geographic
        in nature (as determined by the `units` metadata attribute).

        Not all variables have dimensions, which necessitates a check on
        their existence before determining the dimension is geographic.

        """
        return set(
            dimension
            for dimension in self.get_required_dimensions(variables)
            if self.get_variable(dimension).is_geographic()
        )

    def get_projected_spatial_dimensions(self, variables: set[str]) -> set[str]:
        """Return a single set of all the variables that are both used as
        dimensions for any of the input variables, and that are projected
        in nature (as determined by the `standard_name` metadata
        attribute).

        """
        return set(
            dimension
            for dimension in self.get_required_dimensions(variables)
            if self.get_variable(dimension).is_projection_x_or_y()
        )

    def get_temporal_dimensions(self, variables: set[str]) -> set[str]:
        """Return a single set of all variables that are both used as
        dimensions for any of the input variables, and that are temporal
        in nature (as determined by the `units` metadata attribute).

        Not all variables have dimensions, which necessitates a check on
        their existence before determining the dimension is temporal.

        """
        return set(
            dimension
            for dimension in self.get_required_dimensions(variables)
            if self.get_variable(dimension).is_temporal()
        )

    def get_variables_with_dimensions(self, dimensions: set[str]) -> set[str]:
        """Return a single set of all variables that include all the supplied
        dimensions as a subset of their own dimensions.

        """
        return set(
            variable
            for variable in self.get_all_variables()
            if dimensions.issubset(set(self.get_variable(variable).dimensions))
        )

    def group_variables_by_dimensions(self) -> DimensionsGroupType:
        """Retrieve a dictionary that groups all variables in a file by the
        dimensions for their arrays. Example output for M2I3NPASM:

        ```
        {
            ('/time', '/lev', '/lat', '/lon'): {'/EPV', '/H', ...},
            ('/time', '/lat', '/lon'): {'/PHIS', '/PS', '/SLP'},
            ('/lev', ): {'/lev', },
            ('/lat', ): {'/lat', },
            ('/lon', ): {'/lon', },
            ('/time', ): {'/time', },
        }
        ```

        """
        grouped_variables = {}

        for variable_name in self.get_all_variables():
            variable = self.get_variable(variable_name)
            variable_dimensions = tuple(variable.dimensions)

            if variable_dimensions in grouped_variables:
                grouped_variables[variable_dimensions].add(variable_name)
            else:
                grouped_variables[variable_dimensions] = {
                    variable_name,
                }

        return grouped_variables

    def group_variables_by_horizontal_dimensions(self) -> DimensionsGroupType:
        """Retrieve a dictionary that groups all variables by shared
        horizontal spatial dimensions (e.g., (lon, lat) or (x, y)),
        regardless of other dimensions. This will, for example, group
        variables with dimensions (time, lat, lon) with variables only
        having dimensions (lat, lon). Note, though, the ordering is
        considered, so variables with dimensions (lat, lon) will not be
        grouped with variables having dimensions (lon, lat).

        """
        grid_groups = self.group_variables_by_dimensions()

        horizontal_groups = {}

        for grid_dimensions, variables in grid_groups.items():
            horizontal_dimensions = tuple(
                dimension
                for dimension in grid_dimensions
                if (
                    self.get_variable(dimension) is not None
                    and (
                        self.get_variable(dimension).is_geographic()
                        or self.get_variable(dimension).is_projection_x_or_y()
                    )
                )
            )

            if horizontal_dimensions in horizontal_groups:
                horizontal_groups[horizontal_dimensions].update(variables)
            else:
                horizontal_groups[horizontal_dimensions] = variables

        return horizontal_groups

    @staticmethod
    def exclude_fake_dimensions(variable_set: set[str]) -> set[str]:
        """An OPeNDAP `.dmr` can contain fake dimensions, used to supplement
        missing information for a granule. These cannot be retrieved when
        requesting a subset from an OPeNDAP server, and must be removed
        from the list of required variables.

        """
        fakedim_pattern = re.compile(r'.*/FakeDim\d+')

        return {
            variable for variable in variable_set if not fakedim_pattern.match(variable)
        }


class VarInfoFromDmr(VarInfoBase):
    """A child class that inherits from `VarInfoBase` and implements functions
    to retrieve a dataset from a `.dmr` file, and the extract variables
    from the resulting XML tree.

    """

    def _read_dataset(self, file_path: str):
        """Extract the XML tree and namespace from an OPeNDAP `.dmr` file."""
        with open(file_path, 'r', encoding='utf-8') as file_handler:
            dmr_content = file_handler.read()

        self.dataset = ET.fromstring(dmr_content)
        self.namespace = get_xml_namespace(self.dataset)

    def _set_short_name(self):
        """Iterate through all suggested locations for the collection short
        name, as listed in the configuration file. For each location, perform a
        search for an XML element in the DMR document for that element and, if
        found, retrieve the value of that element.

        """

        self.short_name = next(
            (
                get_full_path_xml_attribute(
                    self.dataset, short_name_path, self.namespace
                )
                for short_name_path in self.var_info_config.get(
                    'CollectionShortNamePath', []
                )
                if get_full_path_xml_attribute(
                    self.dataset, short_name_path, self.namespace
                )
                is not None
            ),
            None,
        )

    def _extract_variables(self):
        """Iterate through all children of the `.dmr` root dataset element.
        If the child matches one of the DAP4 variable types, then create an
        instance of the `VariableFromDmr` class, and assign it to either
        the `variables_with_coordinates` or the `metadata_variables`
        dictionary accordingly.

        """

        def save_variable(output, group_path, element):
            element_path = '/'.join([group_path, element.get('name')])
            variable = VariableFromDmr(
                element,
                self.cf_config,
                self.namespace,
                element_path,
                self.all_dimensions_sizes,
            )
            output[variable.full_name_path] = variable
            self._assign_variable(variable)

        self.find_all_dimensions_sizes(
            self.dataset,
            set(DAP4_TO_NUMPY_MAP.keys()),
            '/',
        )

        all_variables = {}

        self.traverse_elements(
            self.dataset,
            set(DAP4_TO_NUMPY_MAP.keys()),
            save_variable,
            all_variables,
            '/',
        )

        self._remove_non_variable_references()

    def _remove_non_variable_references(self):
        """After all references have been combined, remove those that point to
        non-existent variables. For example dimensions that are present in
        a variable to only denote array size in that dimension. This must
        be done after all variables are parsed, to ensure a reference isn't
        being made to a variable that hasn't yet been extracted.

        """
        self.references = {
            reference
            for reference in self.references
            if self.get_variable(reference) is not None
        }

    def find_all_dimensions_sizes(
        self,
        element: ET.Element,
        element_types: set[str],
        group_path: str,
    ):
        """Perform a depth first search of the `.dmr` `Dimension` element
        before parsing the variables. This ensures that any variable can
        reference a dimension defined later in the dataset. Once an element
        is found, it is then assigned to the `all_dimensions_sizes{}` dictionary.

        """
        group_path = group_path.rstrip('/')

        for child in list(element):
            # Ff it is a Dimension, assign dictionary all_dimensions_sizes{}
            # else, if it is a Group, call this function again
            element_type = child.tag.replace(self.namespace, '')

            if element_type == 'Dimension':
                if child.attrib['size'] is not None:
                    dim_name = '/'.join([group_path, child.get('name')])
                    self.all_dimensions_sizes[dim_name] = int(child.attrib['size'])
            elif element_type == 'Group':
                new_group_path = '/'.join([group_path, child.get('name')])

                self.find_all_dimensions_sizes(child, element_types, new_group_path)

    def traverse_elements(
        self,
        element: ET.Element,
        element_types: set[str],
        operation,
        output,
        group_path: str,
    ):
        """Perform a depth first search of the `.dmr` `Dataset` element.
        When a variable is located perform an operation on the supplied
        output object, using the supplied function or class.

        """
        self.groups[group_path] = GroupFromDmr(
            element,
            self.cf_config,
            namespace=self.namespace,
            full_name_path=group_path,
        )

        group_path = group_path.rstrip('/')

        for child in list(element):
            # If it is in the DAP4 list: use the function
            # else, if it is a Group, assign to dictionary and call this
            # function again
            element_type = child.tag.replace(self.namespace, '')

            if element_type in element_types:
                operation(output, group_path, child)
            elif element_type == 'Group':
                new_group_path = '/'.join([group_path, child.get('name')])

                self.traverse_elements(
                    child, element_types, operation, output, new_group_path
                )


class VarInfoFromNetCDF4(VarInfoBase):
    """A child class that inherits from `VarInfoBase` and implements functions
    to retrieve a dataset from a NetCDF-4 file, and extract the variables
    by traversing the granule structure.

    """

    def _read_dataset(self, file_path: str):
        """Set the dataset to the file path for the NetCDF-4 file. This is
        done instead of assigning a `netCDF4.Dataset` instance, so that the
        file is not still in memory after being parsed, so that other
        services can interact with the NetCDF-4 file without any conflicts.

        """
        self.dataset = file_path

    def _set_short_name(self):
        """Iterate through all suggested locations for the collection short
        name, as listed in the configuration file. For each location, perform a
        search for a metadata attribute in the NetCDF-4 file for that attribute
        and, if found, retrieve the value of that attribute.

        """

        with Dataset(self.dataset, 'r') as dataset:
            self.short_name = next(
                (
                    get_full_path_netcdf4_attribute(dataset, short_name_path)
                    for short_name_path in self.var_info_config.get(
                        'CollectionShortNamePath', []
                    )
                    if get_full_path_netcdf4_attribute(dataset, short_name_path)
                    is not None
                ),
                None,
            )

    def _extract_variables(self):
        """Traverse all groups of the NetCDF-4 file, beginning at the  root
        group.

        """
        with Dataset(self.dataset, 'r') as dataset:
            self._parse_group(dataset)

    def _parse_group(self, group: Dataset | Group):
        """If the child matches one of the DAP4 variable types, then create an
        instance of the `VariableFromNetCDF4` class, and assign it to either
        the `variables_with_coordinates` or the `metadata_variables`
        dictionary accordingly. Child groups are added to the `groups`
        dictionary under the fully resolved path to that group.

        """
        self.groups[group.path] = GroupFromNetCDF4(
            group,
            self.cf_config,
            namespace=self.namespace,
            full_name_path=group.path,
        )

        for netcdf4_variable in group.variables.values():
            variable_path = '/'.join([group.path, netcdf4_variable.name])
            variable_path = f'/{variable_path.lstrip("/")}'

            variable = VariableFromNetCDF4(
                netcdf4_variable,
                self.cf_config,
                namespace=self.namespace,
                full_name_path=variable_path,
            )

            for index, dimensions_name in enumerate(variable.dimensions):
                self.all_dimensions_sizes[dimensions_name] = variable.shape[index]

            self._assign_variable(variable)

        for child_group in group.groups.values():
            self._parse_group(child_group)



"""This module generates UMM-Var records for a specified collection, and
optionally will publish those records to CMR.
`generation_collection_umm_var` will:

* Performing a CMR granule search with collection-based query parameters.
* Downloading the granule locally.
* Parse the variable metadata from the granule using VarInfoFromNetCDF4.
* Generate UMM-Var JSON as a dictionary for each identified variable.
* (Optional) Publish each UMM-Var entry to the selected CMR environment.

"""

from __future__ import annotations

from tempfile import TemporaryDirectory
from typing import Union
import re

from cmr import CMR_UAT

from varinfo import VarInfoFromNetCDF4, VarInfoFromDmr
from varinfo.cmr_search import (
    CmrEnvType,
    download_granule,
    get_granule_link,
    get_granules,
    get_edl_token_header,
    get_dmr_xml_url,
)
from varinfo.umm_var import get_all_umm_var, publish_all_umm_var


# Custom return type: either a list of UMM-Var JSON (a list of dictionaries),
# or a list of strings (either concept IDs or error strings).
UmmVarReturnType = list[Union[dict, str]]


def generate_collection_umm_var(
    collection_concept_id: str,
    auth_header: str,
    cmr_env: CmrEnvType = CMR_UAT,
    publish: bool = False,
    config_file: str | None = None,
    use_dmr: bool = False,
) -> UmmVarReturnType:
    """Run all the of the functions for downloading and publishing
    a UMM-Var entry to CMR given:

    * collection_concept_id: Concept ID for collection that variables have
      been generated for.
    * cmr_env: URLs for CMR environments (OPS, UAT, and SIT)
    * auth_header: Authorization HTTP header, containing a LaunchPad
      token: 'Authorization: <token>'
    * publish: Optional argument determining whether to publish the
      generated UMM-Var records to the indicated CMR instance. Defaults to
      False.
    * config_file: Optional argument to provide a configuration file that
      could be used to override any known errors in a collection. Defaults to None
    Note - if attempting to publish to CMR, a LaunchPad token must be used.

    """
    # Get an EDL token and its header given a LaunchPad token
    auth_header_edl_token = get_edl_token_header(auth_header, cmr_env)

    granule_response = get_granules(
        collection_concept_id, cmr_env=cmr_env, auth_header=auth_header_edl_token
    )

    if use_dmr:
        # Get OPeNDAP data URL with `.dml.xml` appended
        granule_link = get_dmr_xml_url(granule_response)

    else:
        # Get the data download URL for the most recent granule (NetCDF-4 file)
        granule_link = get_granule_link(granule_response)

    with TemporaryDirectory() as temp_dir:
        # Download file to runtime environment
        local_granule = download_granule(
            granule_link, auth_header_edl_token, out_directory=temp_dir
        )

        if use_dmr:
            # Parse the granule with VarInfoFromDmr to map all variables and relations:
            var_info = VarInfoFromDmr(local_granule, config_file=config_file)
        else:
            # Parse the granule with VarInfoFromNetCDF4 to map all variables and relations:
            var_info = VarInfoFromNetCDF4(local_granule, config_file=config_file)

    # Generate all the UMM-Var records:
    all_umm_var_records = get_all_umm_var(var_info)

    if publish:
        # Publish to CMR and construct an output object that is a list of
        # strings. These strings will be either variable concept IDs or
        # error messages returned from CMR.
        publication_response = publish_all_umm_var(
            collection_concept_id, all_umm_var_records, auth_header, cmr_env
        )

        # Produce a list indicating publication information for all variables.
        # Variables that were successfully published will have a list element
        # providing their variable concept ID. Any variables that were
        # not successfully published will instead have an element containing
        # the variable name and the CMR error (e.g., 'variable: CMR error...').
        return_value = [
            (
                variable_response
                if is_variable_concept_id(variable_response)
                else ': '.join([variable_name, variable_response])
            )
            for variable_name, variable_response in publication_response.items()
        ]
    else:
        # If not publishing, return the full UMM-Var JSON records
        return_value = list(all_umm_var_records.values())

    return return_value


def is_variable_concept_id(possible_concept_id: str) -> bool:
    """A helper function to identify if a given string conforms to the
    expected structure of a variable concept ID, e.g., 'V1234567890-PROV'.

    """
    return bool(re.match(r'^V\d{10}-\w+$', possible_concept_id))
